# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""CLI for requirements builder."""

from __future__ import absolute_import, print_function

import click

from .requirements_builder import iter_requirements


@click.command()
@click.option(
    '--level',
    '-l',
    default='pypi',
    type=click.Choice(['min', 'pypi', 'dev']),
    help='Specifies desired requirements level.'
    '"min" requests the minimal requirement that is specified, '
    '"pypi" requests the maximum version that satisfies the '
    'constrains and is available in PyPi. '
    '"dev" includes experimental developer versions for VCSs.'
)
@click.option(
    '--extras',
    '-e',
    default='',
    help='Comma separated list of extras.',
    multiple=True
)
@click.option(
    '--req',
    '-r',
    default=None,
    help='Requirements file.',
    metavar='PATH',
    type=click.Path(
        exists=True, dir_okay=False, readable=True, resolve_path=True
    )
)
@click.option(
    '--output', '-o', default='-', help='Output file.', type=click.File('w')
)
@click.argument('setup', type=click.File('r'), required=False)
@click.pass_context
def cli(ctx, level, extras, req, output, setup):
    """Calculate requirements for different purposes."""
    if level == 'dev' and not req:
        raise click.UsageError(
            "You must specify --req when using 'dev' level."
        )
    if not setup and not req:
        raise click.MissingParameter(
            ctx=ctx, param_hint='"setup"', param_type='argument'
        )
    extras = set(req.strip() for extra in extras for req in extra.split(','))

    lines = (
        '{0}\n'.format(req)
        for req in iter_requirements(level, extras, req, setup)
    )
    output.writelines(lines)



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Tests for `requirements-builder` module."""

from os.path import abspath, dirname, join

from requirements_builder import __version__, iter_requirements

REQ = abspath(join(dirname(__file__), "../requirements.devel.txt"))
SETUP = abspath(join(dirname(__file__), "../setup.py"))


def test_version():
    """Test requirements-builder."""
    assert __version__


def test_iter_requirements():
    """Test requirements-builder."""
    # Min
    with open(SETUP) as f:
        assert list(iter_requirements("min", [], '', f)) == \
            ['click==6.1.0', 'mock==1.3.0']

    # PyPI
    with open(SETUP) as f:
        assert list(iter_requirements("pypi", [], '', f)) == \
            ['click>=6.1.0', 'mock>=1.3.0']

    # Dev
    with open(SETUP) as f:
        assert list(iter_requirements("dev", [], REQ, f)) == \
            ['-e git+https://github.com/pallets/click.git#egg=click',
             'mock>=1.3.0']



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017, 2018 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Build requirements files from setup.py requirements."""

from setuptools import setup

import os
import testpkh

dirname = os.path.dirname(__file__)

requirements = [
    'click>=5.0.0',
    'mock>=1.3.0',
    'CairoSVG<2.0.0,>=1.0.20',
    'functools32>=3.2.3-2; python_version=="2.7"',
    'invenio-records~=1.0.0',
    'invenio[base,auth,metadata]>=3.0.0',
]

extras_require = {
    'docs': ['Sphinx>=1.4.2'],
    'tests': ['pytest>=2.7'],
    'flask': ['Flask>=0.11'],
    ':python_version=="2.7"': ['ipaddr>=2.1.11']
}

setup(
    name='testpkh',
    version=testpkh.__version__,
    install_requires=requirements,
    extras_require=extras_require,
)



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017, 2018 CERN.
# Copyright (C) 2018 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Tests for CLI module."""

import shutil
import sys
from os import getcwd
from os.path import abspath, dirname, join

import pytest
from click.testing import CliRunner
from requirements_builder.cli import cli

DATA = abspath(join(dirname(__file__), 'data/'))


@pytest.fixture(scope='function')
def runner():
    """Click test runner."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copytree(DATA, abspath(join(getcwd(), 'data/')))
        yield runner


def test_cli_min(runner):
    """Test cli."""
    result = runner.invoke(cli, ['-l', 'min', 'data/setup.py'])
    assert result.exit_code == 0
    if sys.version_info[:2] == (2, 7):
        assert result.output == \
            'CairoSVG==1.0.20\n' \
            'click==5.0.0\n' \
            'functools32==3.2.3-2\n' \
            'invenio[auth,base,metadata]==3.0.0\n' \
            'invenio-records==1.0.0\n' \
            'ipaddr==2.1.11\n' \
            'mock==1.3.0\n'
    else:
        assert result.output == \
            'CairoSVG==1.0.20\n' \
            'click==5.0.0\n' \
            'invenio[auth,base,metadata]==3.0.0\n' \
            'invenio-records==1.0.0\n' \
            'mock==1.3.0\n'


def test_cli_pypi(runner):
    """Test cli."""
    result = runner.invoke(cli, ['-l', 'pypi', 'data/setup.py'])
    assert result.exit_code == 0

    if sys.version_info[:2] == (2, 7):
        assert result.output == \
            'CairoSVG<2.0.0,>=1.0.20\n' \
            'click>=5.0.0\n' \
            'functools32>=3.2.3-2\n' \
            'invenio[auth,base,metadata]>=3.0.0\n' \
            'invenio-records>=1.0.0,==1.0.*\n' \
            'ipaddr>=2.1.11\n' \
            'mock>=1.3.0\n'
    else:
        assert result.output == \
            'CairoSVG<2.0.0,>=1.0.20\n' \
            'click>=5.0.0\n' \
            'invenio[auth,base,metadata]>=3.0.0\n' \
            'invenio-records>=1.0.0,==1.0.*\n' \
            'mock>=1.3.0\n'


def test_cli_dev_erro(runner):
    """Test cli."""
    result = runner.invoke(cli, ['-l', 'dev', 'data/setup.py'])
    assert result.exit_code == 2


def test_cli_dev(runner):
    """Test cli."""
    result = runner.invoke(
        cli, ['-l', 'dev', '-r', 'data/req.txt', 'data/setup.py']
    )
    assert result.exit_code == 0
    if sys.version_info[:2] == (2, 7):
        assert result.output == \
            'CairoSVG<2.0.0,>=1.0.20\n' \
            '-e git+https://github.com/mitsuhiko/click.git#egg=click\n' \
            'Cython>=0.20\n' \
            'functools32>=3.2.3-2\n' \
            'invenio[auth,base,metadata]>=3.0.0\n' \
            'invenio-records>=1.0.0,==1.0.*\n' \
            'ipaddr>=2.1.11\n' \
            'mock>=1.3.0\n'
    else:
        assert result.output == \
            'CairoSVG<2.0.0,>=1.0.20\n' \
            '-e git+https://github.com/mitsuhiko/click.git#egg=click\n' \
            'Cython>=0.20\n' \
            'invenio[auth,base,metadata]>=3.0.0\n' \
            'invenio-records>=1.0.0,==1.0.*\n' \
            'mock>=1.3.0\n'


def test_cli_min_output(runner):
    """Test cli."""
    result = runner.invoke(
        cli, ['-l', 'min', '-o', 'requirements.txt', 'data/setup.py']
    )
    assert result.exit_code == 0
    assert result.output == ''
    with open(join(getcwd(), 'requirements.txt')) as f:
        if sys.version_info[:2] == (2, 7):
            assert f.read() == \
                'CairoSVG==1.0.20\n' \
                'click==5.0.0\n' \
                'functools32==3.2.3-2\n' \
                'invenio[auth,base,metadata]==3.0.0\n' \
                'invenio-records==1.0.0\n' \
                'ipaddr==2.1.11\n' \
                'mock==1.3.0\n'
        else:
            assert f.read() == \
                'CairoSVG==1.0.20\n' \
                'click==5.0.0\n' \
                'invenio[auth,base,metadata]==3.0.0\n' \
                'invenio-records==1.0.0\n' \
                'mock==1.3.0\n'


def test_cli_extras(runner):
    """Test cli option extras."""
    output = ['CairoSVG==1.0.20',
              'click==5.0.0',
              'invenio[auth,base,metadata]==3.0.0',
              'invenio-records==1.0.0',
              'mock==1.3.0']
    if sys.version_info[:2] == (2, 7):
        output.append('functools32==3.2.3-2')
        output.append('ipaddr==2.1.11')

    result = runner.invoke(
        cli, ['-l', 'min', '-e', 'docs', 'data/setup.py']
    )
    assert result.exit_code == 0
    assert set(result.output.split('\n')
               ) == set(output + ['Sphinx==1.4.2', ''])

    result = runner.invoke(
        cli, ['-l', 'min', '-e', 'docs, tests', 'data/setup.py']
    )
    assert result.exit_code == 0
    assert set(result.output.split('\n')
               ) == set(output + ['pytest==2.7', 'Sphinx==1.4.2', ''])

    result = runner.invoke(
        cli,
        ['-l', 'min', '-e', 'docs, tests', '-e', 'flask', 'data/setup.py']
    )
    assert result.exit_code == 0
    assert set(
        result.output.split('\n')
    ) == set(output + ['pytest==2.7', 'Sphinx==1.4.2', 'Flask==0.11', ''])


def test_cli_no_setup(runner):
    """Test cli with no setup.py provided."""
    with runner.isolated_filesystem():
        result = runner.invoke(cli, ['-l', 'min'])

        assert result.exit_code != 0
        assert (
            result.output.strip().split('\n')[-1] ==
            'Error: Missing argument "setup".'
        )


def test_cli_with_only_requirements(runner):
    """Test cli with no setup.py provided but a req.txt"""
    with runner.isolated_filesystem():
        shutil.copytree(DATA, abspath(join(getcwd(), 'data/')))

        result = runner.invoke(cli, ['-l', 'min', '-r', 'data/req.txt'])

        assert result.exit_code == 0
        assert set(result.output.strip().split('\n')) == set([
            'Cython==0.20',
            '-e git+https://github.com/mitsuhiko/click.git#egg=click'
        ]), result.output.split('\n')


def test_cli_setup_with_if_main(runner):
    """Test loading setup.py with 'if __name__ == '__main__' around setup."""
    with runner.isolated_filesystem():
        shutil.copytree(DATA, abspath(join(getcwd(), 'data/')))
        result = runner.invoke(cli, ['-l', 'min', 'data/setup_if_main.py'])

        assert result.exit_code == 0
        assert result.output == 'click==5.0.0\n'



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2017 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Build requirements files from setup.py requirements.

.. code-block:: console

   $ requirements-builder --help
   Usage: requirements-builder [OPTIONS] SETUP

     Calculate requirements for different purposes.

   Options:
     -l, --level [min|pypi|dev]  Specifies desired requirements level."min"
                                 requests the minimal requirement that is
                                 specified, "pypi" requests the maximum version
                                 that satisfies the constrains and is available
                                 in PyPi. "dev" includes experimental developer
                                 versions for VCSs.
     -e, --extras TEXT           Comma separated list of extras.
     -r, --req PATH              Requirements file.
     --help                      Show this message and exit.

TravisCI
--------
Following is an example of how to integrate Requirements-Builder with
TravisCI:

.. code-block:: yaml

   env:
     - REQUIREMENTS=lowest
     - REQUIREMENTS=release
     - REQUIREMENTS=devel

   python:
     - "2.7"
     - "3.3"
     - "3.4"
     - "3.5"

   before_install:
     - "travis_retry pip install --upgrade pip"
     - "travis_retry pip install requirements-builder"
     - "requirements-builder --level=min setup.py
        > .travis-lowest-requirements.txt"
     - "requirements-builder --level=pypi setup.py
        > .travis-release-requirements.txt"
     - "requirements-builder --level=dev --req requirements-devel.txt setup.py
        > .travis-devel-requirements.txt"

   install:
     - "travis_retry pip install -r .travis-$REQUIREMENTS-requirements.txt"
     - "pip install -e ."

Tox
---
How to integrate Requirements-Builder with tox:

.. code-block:: ini

    [tox]
    skipdist = True

    envlist =
        {py27}-{min,pypi,dev}
        {py34}-{min,pypi,dev}
        {py35}-{min,pypi,dev}
        {py36}-{min,pypi,dev}

    [testenv]
    setenv =
        PYTHONPATH = {toxinidir}:{toxinidir}

    deps =
        Requirements-Builder

    commands =
        min: requirements-builder --level=min -o .tox/min.txt setup.py
        min: pip install -r .tox/min.txt
        pypi: requirements-builder --level=min -o .tox/pypi.txt setup.py
        pypi: pip install -r .tox/pypi.txt
        dev: requirements-builder -l dev -r devel.txt -o .tox/dev.txt setup.py
        dev: pip install -r .tox/requirements-dev.txt
        pip install -e .
        {envpython} setup.py test # or: py.test, nose, coverage run, etc.

Tox-travis
^^^^^^^^^^
When you are using tox, the TravisCI setup becomes much simpler.

.. code-block:: yaml

    language: python

    python:
      - "2.7"
      - "3.4"
      - "3.5"
      - "3.6"

    install:
      - pip install tox tox-travis

    script:
      - tox
"""

from __future__ import absolute_import, print_function

from .requirements_builder import iter_requirements
from .version import __version__

__all__ = ('__version__', )



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2017 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Build requirements files from setup.py requirements."""

from setuptools import setup

import testpkh


requirements = [
    'click>=5.0.0',
]

extras_require = {
    'docs': ['Sphinx>=1.4.2'],
}

if __name__ == "__main__":
    setup(
        name='testpkh',
        version=testpkh.__version__,
        install_requires=requirements,
        extras_require=extras_require,
    )



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Build requirements files from setup.py requirements."""

import os

from setuptools import setup

# Get the version string.  Cannot be done with import!
g = {}
with open(os.path.join('requirements_builder', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGES.rst') as history_file:
    history = history_file.read().replace('.. :changes:', '')

install_requires = [
    'click>=6.1.0',
    'mock>=1.3.0',
]

tests_require = [
    'check-manifest>=0.25',
    'coverage>=4.0',
    'isort>=4.0.0',
    'pydocstyle>=1.0.0',
    'pytest-cache>=1.0',
    'pytest-cov>=2.0.0',
    'pytest-pep8>=1.0.6',
    'pytest>=2.8.0',
]

extras_require = {
    'docs': [
        'Sphinx<1.5.0,>=1.4.2',
        'docutils<0.13,>=0.12',
    ],
    'tests': tests_require,
}

extras_require['all'] = extras_require['tests'] + extras_require['docs']

setup_requires = ['pytest-runner>=2.6.2', ]

setup(
    name='requirements-builder',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    author="Invenio Collaboration",
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/requirements-builder',
    entry_points={
        'console_scripts':
            ["requirements-builder = requirements_builder.cli:cli"]
    },
    packages=['requirements_builder', ],
    include_package_data=True,
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    license='BSD',
    zip_safe=False,
    keywords='requirements-builder',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017, 2018 CERN.
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Generate requirements from `setup.py` and `requirements-devel.txt`."""

from __future__ import absolute_import, print_function

import os
import re
import sys

import mock
import pkg_resources
import setuptools


def parse_set(string):
    """Parse set from comma separated string."""
    string = string.strip()
    if string:
        return set(string.split(","))
    else:
        return set()


def minver_error(pkg_name):
    """Report error about missing minimum version constraint and exit."""
    print(
        'ERROR: specify minimal version of "{0}" using '
        '">=" or "=="'.format(pkg_name),
        file=sys.stderr
    )
    sys.exit(1)


def build_pkg_name(pkg):
    """Build package name, including extras if present."""
    if pkg.extras:
        return '{0}[{1}]'.format(
            pkg.project_name, ','.join(sorted(pkg.extras)))
    return pkg.project_name


def parse_pip_file(path):
    """Parse pip requirements file."""
    # requirement lines sorted by importance
    # also collect other pip commands
    rdev = {}
    rnormal = []
    stuff = []

    try:
        with open(path) as f:
            for line in f:
                line = line.strip()

                # see https://pip.readthedocs.io/en/1.1/requirements.html
                if line.startswith('-e'):
                    # devel requirement
                    splitted = line.split('#egg=')
                    rdev[splitted[1].lower()] = line

                elif line.startswith('-r'):
                    # recursive file command
                    splitted = re.split('-r\\s+', line)
                    subrdev, subrnormal, substuff = parse_pip_file(
                        os.path.join(os.path.dirname(path), splitted[1])
                    )
                    for k, v in subrdev.items():
                        if k not in rdev:
                            rdev[k] = v
                    rnormal.extend(subrnormal)
                elif line.startswith('-'):
                    # another special command we don't recognize
                    stuff.append(line)
                else:
                    # ordinary requirement, similarly to them used in setup.py
                    rnormal.append(line)
    except IOError:
        print(
            'Warning: could not parse requirements file "{0}"!'.format(path),
            file=sys.stderr
        )

    return rdev, rnormal, stuff


def iter_requirements(level, extras, pip_file, setup_fp):
    """Iterate over requirements."""
    result = dict()
    requires = []
    stuff = []
    if level == 'dev' or setup_fp is None:
        result, requires, stuff = parse_pip_file(pip_file)

    install_requires = []
    requires_extras = {}
    if setup_fp is not None:
        with mock.patch.object(setuptools, 'setup') as mock_setup:
            sys.path.append(os.path.dirname(setup_fp.name))
            g = {'__file__': setup_fp.name, '__name__': '__main__'}
            exec(setup_fp.read(), g)
            sys.path.pop()
            assert g['setup']  # silence warning about unused imports

        # called arguments are in `mock_setup.call_args`
        mock_args, mock_kwargs = mock_setup.call_args
        install_requires = mock_kwargs.get(
            'install_requires', install_requires
        )
        requires_extras = mock_kwargs.get('extras_require', requires_extras)

    install_requires.extend(requires)

    for e, reqs in requires_extras.items():
        # Handle conditions on extras. See pkginfo_to_metadata function
        # in Wheel for details.
        condition = ''
        if ':' in e:
            e, condition = e.split(':', 1)
        if not e or e in extras:
            if condition:
                reqs = ['{0}; {1}'.format(r, condition) for r in reqs]
            install_requires.extend(reqs)

    for pkg in pkg_resources.parse_requirements(install_requires):
        # skip things we already know
        # FIXME be smarter about merging things

        # Evaluate environment markers skip if not applicable
        if hasattr(pkg, 'marker') and pkg.marker is not None:
            if not pkg.marker.evaluate():
                continue
            else:
                # Remove markers from the output
                pkg.marker = None

        if pkg.key in result:
            continue

        specs = dict(pkg.specs)
        if (('>=' in specs) and ('>' in specs)) \
                or (('<=' in specs) and ('<' in specs)):
            print(
                'ERROR: Do not specify such weird constraints! '
                '("{0}")'.format(pkg),
                file=sys.stderr
            )
            sys.exit(1)

        if '==' in specs:
            result[pkg.key] = '{0}=={1}'.format(
                build_pkg_name(pkg), specs['=='])

        elif '>=' in specs:
            if level == 'min':
                result[pkg.key] = '{0}=={1}'.format(
                    build_pkg_name(pkg), specs['>=']
                )
            else:
                result[pkg.key] = pkg

        elif '>' in specs:
            if level == 'min':
                minver_error(build_pkg_name(pkg))
            else:
                result[pkg.key] = pkg

        elif '~=' in specs:
            if level == 'min':
                result[pkg.key] = '{0}=={1}'.format(
                    build_pkg_name(pkg), specs['~='])
            else:
                ver, _ = os.path.splitext(specs['~='])
                result[pkg.key] = '{0}>={1},=={2}.*'.format(
                    build_pkg_name(pkg), specs['~='], ver)

        else:
            if level == 'min':
                minver_error(build_pkg_name(pkg))
            else:
                result[pkg.key] = build_pkg_name(pkg)

    for s in stuff:
        yield s

    for k in sorted(result.keys()):
        yield str(result[k])



# -*- coding: utf-8 -*-
#
# This file is part of Requirements-Builder
# Copyright (C) 2015, 2016, 2017 CERN.
# Copyright (C) 2018 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Requirements-Builder is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.
#
"""Version information for Requirements-Builder.

This file is imported by ``requirements-builder.__init__``, and parsed by
``setup.py`` as well as ``docs/conf.py``.
"""

# Do not change the format of this next line. Doing so risks breaking
# setup.py and docs/conf.py

__version__ = "0.3.0"
